//+------------------------------------------------------------------+
//|                                            CFractalIndexLine.mqh |
//|                                Copyright 2019, Korotchenko Roman |
//|                        https://login.mql5.com/en/users/Solitonic |
//|                                             Revision 14 May 2019 |
//+------------------------------------------------------------------+
#property copyright "Copyright 2019, Korotchenko Roman"
#property link        "https://login.mql5.com/en/users/Solitonic"
#include <FRACTAL\CXY_Array.mqh>
#include <FRACTAL\fConfidenceIntervalMU.mqh>

#define MAX_SCALE_COUNT 128  // max. number of partitions of the analyzed fragment of the series. Power of 2
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class CFractalIndexLine
  {
private:
   int               m_nfrgm;                // size of the allocated class array
   double            LogAmplVariations[]; // natural logarithms of amplitude variations of the function corresponding to diff partition variants LogDeltaScales;
   double            LogDeltaScales[];    // natural logarithms of series dividing scales 

   int  ipow2(const int Num) { double v=log(Num)/log(2.); return((int)round(v)); }

   int               LoadFragment(const double &series[],const int N0,const int N1,const double hourDuration);
   double            CalcAmplVariation(const int NumPartDivide,int idxAmplVar=-1);
   double            BoundaryArray(const int NumPartDivide,const double &x[],const int leftIdx,const int rightIdx,const double hSampling,CXY_Array &bnd,int &nIntervalPoints);

public:
   int               Nfrgm;              // length of the time series fragment
   double            hSampling;       // discretization of time series between points on hours

   double            fragment[];      // fragment of the time series being processed
   CXY_Array        *Boundaries;  // to calculate boundaries of segmentation within the time series fragment
   CXY_Array        *DblLog;      // to calculate V(delta) in double logarithmic scale in the time series fragment

   double            Mu,Hurst;       // Mu - local fractal index, Hurst - Hurst exponent
   double            ConfInterval[2]; // 95% confidence interval for the fractal index (left and right boundaries)
   double            R2det,KR2;      // R2det - determination coefficient (stability of determination of fractal index.) KR2 = 1-R2det "oscillation" of stability 

                     CFractalIndexLine();
                    ~CFractalIndexLine();

   double            CalculateFractalIndex(const double &series[],const int N0,const int N1,const double hourSampling,int CountFragmentScale=0);

  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
CFractalIndexLine::CFractalIndexLine()
  {
   Boundaries = new CXY_Array(MAX_SCALE_COUNT);
   DblLog     = new CXY_Array(MAX_SCALE_COUNT);

   ArrayResize(LogAmplVariations,MAX_SCALE_COUNT); // function's amplitude variations corresponding to diff partition variants LogDeltaScales;
   ArrayResize(LogDeltaScales,MAX_SCALE_COUNT);    // series partitioning scales 

   m_nfrgm=Nfrgm=0;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
CFractalIndexLine::~CFractalIndexLine()
  {
   ArrayFree(fragment);
   ArrayFree(LogAmplVariations);
   ArrayFree(LogDeltaScales);
   delete Boundaries;
   delete DblLog;
  }
//+------------------------------------------------------------------+

double CFractalIndexLine::CalculateFractalIndex(const double &series[],const int N0,const int N1,
                                                const double hourSampling,int CountFragmentScale=0)
  {
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// series[] - time series
// N0, N1 - the left and right boundary points of the series[] array fragment, based on which the fractal index will be estimated
// hourSampling - discretization between points in HOURS
// CountFragmentScale - the number of requested scales to form a set of points, for which the fractal index is calculated
//
// RESULT
// the fractal index (Mu), the Hurst index (Hurst), the Confidence interval 95% (ConfInterval[2], 
// coefficient of determination (R2det) - the closer to 1, the more accurately the calculation points fall on the approximating line
// determining stability for coefficient KR2 = 1-R2det. The closer to zero, the more accurate the calculated value of Mu
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// 1. Load the internal fragment with values from the time series     
   LoadFragment(series,N0,N1,hourSampling);

// 2. Determine the number of cycles to determine the points of the approximating line
   int   nn2 = (int)floor(Nfrgm/2);  // Partition limits - no less than two points
   int npow2 = (int)ipow2(nn2);      // The number of the powers of two in the Possible partitioning limit;

   if(CountFragmentScale==0) CountFragmentScale=npow2; // default

   int Count=fmin(CountFragmentScale,npow2);                 // limiting the number of variants of series fragment partitioning
   int NumPartDivide;

   for(int i=0; i<=Count; i++)
     {
      NumPartDivide = (int)pow (2,i);      // Number of parts in the series fragment division
      CalcAmplVariation(NumPartDivide, i); // Calculating a point for the approximating line model      
      i=i;
     }

   Mu=fCalculateConfidenceIntervalMU(LogDeltaScales,LogAmplVariations,Count,ConfInterval,R2det);
   Hurst=1-Mu;
   KR2=1-R2det;

   return Mu;
  }
//----------------------------------------------------------------------------------------------------------------------------------

int CFractalIndexLine::LoadFragment(const double &series[],const int N0,const int N1,const double hourSampling)
  {
   int nmin,nmax,nn,k;
   double ScaleAmpl=1;

   hSampling=hourSampling; // "frequency" of discretization = time of "sampling" in hours
   ScaleAmpl=hSampling/24.; // CONVERTING HOURS TO DAY !  

   nmin=(N0<N1)?N0:N1;   // N0 and N1 - boundary values of the fragment
   nmax= (N0<N1)?N1:N0;
   nn=nmax-nmin+1;       // number of points in the loaded fragment

   if(nn>m_nfrgm) { m_nfrgm=ArrayResize(fragment,nn); }
   if(m_nfrgm<=0) return 0; // error of allocation memory

   Nfrgm=nn; k=0;
   for(int i=nmin; i<=nmax; i++)
     {
      fragment[k]=ScaleAmpl*series[i]; k++;  // Scaling to save "SCALE INVARIANCE" of fractal index calculation
     }
// additionally clear calculation arrays for further analysis
   Boundaries.FillByEmpty();
   DblLog.FillByEmpty();
   for(int i=0; i<MAX_SCALE_COUNT; i++)
     {
      LogAmplVariations[i] = EMPTY_VALUE;
      LogDeltaScales[i]    = EMPTY_VALUE;
     }

   return Nfrgm;
  }
//--------------------------------------------------------------------------------------------------------------------------------------

double CFractalIndexLine::CalcAmplVariation(const int NumPartDivide,int idxAmplVar=-1)
  {
// If idxAmplVar=-1, then index in the array is determined automatically (based on the contribution of the power of two in NumPartDivide)
// ALREADY PREVIOUSLY DONE: copying the fragment, setting the time of discretization of the series IN DAYS
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

// 1. DETERMINE THE BORDERS OF INTERVALS CORRESPONDING TO THE SPECIFIC NUMBERS
   int nCheckPoint=0,nIntervalPoints=0; // number of points to check in one division sub-interval 
   double dayDeltaScales=BoundaryArray(NumPartDivide,fragment,0,Nfrgm-1,hSampling,Boundaries,nIntervalPoints);

// 2. GO THROUGH INTERVALS TO DETERMINE LIMIT VALUES OF FUNCTIONS AND OF AMPLITUDE VARIATION
   int countInterval=Boundaries.CountNonEmty();
   int  maxFuncIdx=0,minFuncIdx=0;
   double A,V=0.;

   nCheckPoint=(int)(Boundaries.y[0]-Boundaries.x[0])+1;
   for(int i=0; i<countInterval; i++)
     {
      maxFuncIdx = ArrayMaximum(fragment,(int)Boundaries.x[i],nCheckPoint); // INDEX IN THE ARRAY
      minFuncIdx = ArrayMinimum(fragment,(int)Boundaries.x[i],nCheckPoint);
      A = fragment[maxFuncIdx] - fragment[minFuncIdx];
      V = V+A;
      i=i;
     }

// 3. ACCUMULATION OF RESULTS IN STORAGE    
   if(idxAmplVar==-1) idxAmplVar=ipow2(NumPartDivide); // index in the storage array

   LogDeltaScales   [idxAmplVar] = log(dayDeltaScales); // log-scale of the current division
   LogAmplVariations[idxAmplVar] = log(V);              // log-Amplitude Variation in the specified scale

   return V;
  }
//--------------------------------------------------------------------------------------------------------------------------------------

double CFractalIndexLine::BoundaryArray(const int NumPartDivide,const double &x[],const int leftIdx,const int rightIdx,
                                        const double hourSampling,CXY_Array &bnd,int &nIntervalPoints)
  {
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// NumPartDivide - number of required partitions = NUMBER OF USED ELEMENTS for the ARRAY OF BOUNDARIES: x[],y[]
// x[] - time series
// leftIdx, rightIdx - the left and right boundary points in the x[] array
// hSampling - discretization between points in hours
//
// RESULTS
// bnd[][2] - array of interval borders
//    nIntervalPoints     - number of points in one division sub-interval 
//    bnd[][1] - left boundaries in the sub-interval
//    bnd[][2] - tight boundaries in the sub-interval
//    
// Function result - duration of allocated sub-intervals in time as the number of calculation items (days)x
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   int i,np,minIdx,maxIdx;
   int mb;
   double hDeltaScales; // duration of each of the sub-intervals which will be allocated for a given number of partitions (TIME SCALE)

   if(rightIdx<leftIdx)
     {
      minIdx = rightIdx;
      maxIdx = leftIdx;
     }
   else { minIdx=leftIdx;     maxIdx=rightIdx; } // indexes for a series in a normal order

   np=(maxIdx-minIdx); // Length of the presented segment as of the base in the rectangular covering

                       // parity check
// if( fmod(np,2) != 0.)   np += 1;

   mb=bnd.length;

   if(mb<NumPartDivide) // if necessary, increase the array for boundary points
     {
      bnd.Resize(NumPartDivide,16*NumPartDivide);
     }

   bnd.FillByEmpty();         // EMPTY_VALUE 

   nIntervalPoints=np/NumPartDivide; // number of points in one division sub-interval 

   for(i=0; i<NumPartDivide; i++)
     {
      bnd.x[i] = minIdx + i*nIntervalPoints;   // "left"  point of the sub-interval
      bnd.y[i] = bnd.x[i] + nIntervalPoints;   // "right"  point of the sub-interval    
      i=i;
     }

   hDeltaScales=nIntervalPoints*hourSampling; // Covering scale (length of the base of dividing blocks) IN HOURS
   return hDeltaScales/24.0;  // Covering scale IN DAYS
  }
//-----------------------------------------------------------------------------------------------------------------------------------------------
