//+------------------------------------------------------------------+
//|                                          CConfidenceInterval.mqh |
//|                                Copyright 2019, Korotchenko Roman |
//|                        https://login.mql5.com/en/users/Solitonic |
//|                                             Revision 22 May 2019 |
//+------------------------------------------------------------------+
#property copyright   "Copyright 2019, Roman Korotchenko"
#property link        "https://login.mql5.com/en/users/Solitonic"
#property version   "1.0"
#include <Math\Alglib\specialfunctions.mqh>
//+------------------------------------------------------------------+
//| Calculating the mean and dispersion at given function interval   |
//| INPUT PARAMETERS:                                                |
//|     f[]     -   values                                           |
//|     idx0    -   initial index of the interval                    |
//|     idx1    -   finishing index of the interval                  |
//|                 if indexes are equal to ZERO                     |
//|                 all f[] values are processed                     |
//| OUTPUT PARAMETERS                                                |
//|     Mean    -   average                                          |
//|     Variance-   dispersion                                       |
//+------------------------------------------------------------------+

bool fMeanAndVariance(const double &f[], const int idx0, int idx1, double &mean, double &variance)
{ 
       if(idx1 == 0) idx1 = ArraySize(f)-1;
  int n = idx1 - idx0+1;
  double v1=0;
         
   
   if(n<=0)
      return(false);
//--- Mean
   mean=0;
   for(int i=idx0;i<=idx1;i++)
      mean+=f[i];
   mean/=n;

   variance=0;
   
   //--- Variance 
   if(n!=1)
     {
      //--- calculation
      v1=0;
      for(int i=idx0;i<=idx1;i++)
         v1+=(f[i]-mean)*(f[i]-mean);
      
      variance=v1/n;            
     }
   else
      variance=EMPTY_VALUE;   

 return true;
}
//---------------------------------------------------------------------------------------------


double fLinearApprox(const double &x[], const double &f[], const int idx0, int idx1, double &Coeff, double &B)
{
//+------------------------------------------------------------------+
//| Calculating linear approximation parameters for 2D function (LS) |
//| INPUT PARAMETERS:                                                |
//|     f[]     -   values of ordinates                              |
//|     x[]     -   abscissa values                                  |
//|     idx0    -   initial index of the interval                    |
//|     idx1    -   finishing index of the interval                  |
//|                 if indexes are equal to ZERO                     |
//|                 all set of f[] is processed                      |
//| OUTPUT PARAMETERS                                                |
//|     Coeff   -   coefficient of the line Y=Coeff*X + B            |
//|         B   -   free member                                      |
//|     R2det   -   coefficient of determination                     |
//+------------------------------------------------------------------+

double vv, zz, yavr, xavr, x2avr, yxavr , R2, R2det;
      if(idx1 == 0) idx1 = ArraySize(f)-1;
 int n = idx1 - idx0+1;
 
 yavr = xavr = x2avr = yxavr =0 ;
   
   for ( int i = idx0; i<=idx1; i++)
   {
      xavr = xavr + x[i];
      yavr = yavr + f[i];
      x2avr =  x2avr+ x[i]*x[i];
      yxavr =  yxavr+ f[i]*x[i];
   }
   
   xavr = xavr/n;
   yavr = yavr/n;
  x2avr = x2avr/n;
  yxavr = yxavr/n;
    
      Coeff = (yxavr - xavr*yavr)/(x2avr-xavr*xavr);  // _y = Coeff*x+B;
       B = yavr - Coeff*xavr;
      
   R2 = 0;
   R2det = 0;
   for ( int i = 0; i<n; i++)
   {  
    vv = ( Coeff *x[i] + B - yavr );  R2 = R2 + vv*vv;
    zz = (f[i] - yavr);        R2det = R2det + zz*zz;
   }
    R2det = R2/R2det; // COEFFICIENT OF DETERMINATION

return R2det;
}
//---------------------------------------------------------------------------------------------

double ValueR2det(const double &x[], const double &f[], const int idx0, int idx1,  const double Coeff, const double B)
{
//+------------------------------------------------------------------+
//| Calculate coefficient of determination for linear approximation  |
//| INPUT PARAMETERS:                                                |
//|     f[]     -   function values (ordinates)                      |
//|     x[]     -   abscissa values                                  |
//|     idx0    -   initial index of the interval                    |
//|     idx1    -   finishing index of the interval                  |
//|                 if indexes are equal to ZERO                     |
//|                 all set of f[] is processed                      |
//|  Coeff,B    -   linear regression parameters (Y= Coeff*X+B)      |
//| RESULT                                                           |
//|     R2det   -   coefficient of determination                     |
//+------------------------------------------------------------------+
  double R2, R2det, vv, zz, yavr;
     
      if(idx1 == 0) idx1 = ArraySize(f)-1;
 int n = idx1 - idx0+1;
    
   yavr = 0;   
      for ( int i = idx0; i<=idx1; i++)  yavr +=  f[i];               
   yavr = yavr/n;
   
     R2 = 0;
   R2det = 0;
   for ( int i = 0; i<n; i++)
   {  
    vv = ( Coeff *x[i] + B - yavr );  R2 += (vv*vv);
    zz = (f[i] - yavr);        R2det += (zz*zz);
   }
    R2det = R2/R2det; // COEFFICIENT OF DETERMINATION
 return R2det; 
}
//---------------------------------------------------------------------------------------------

double ConfidenceInterval(const double &x[], const double &f[], const int idx0, int idx1,  const double Coeff, const double B, double alfa = 0.05)
{
//+------------------------------------------------------------------+
//| Calculating deviation to estimate the confidence interval Coeff  |
//| INPUT PARAMETERS:                                                |
//|     f[]     -   function values (ordinates)                      |
//|     x[]     -   abscissa values                                  |
//|     idx0    -   initial index of the interval                    |
//|     idx1    -   finishing index of the interval                  |
//|                 if the index is ZERO                             |
//|                 all f[] values are processed                     |
//|  Coeff,B    -   linear regression parameters (Y= Coeff*X+B)      |
//|     alfa    -   level of significance of the confidence interval |
//|(Value as % of confid.interval = 100*(1-alfa) = 95% default)      |
//|                                                                  |
//| RESULT                                                           |
//|  DeltaCoeff -   deviation for Coeff estimate                     |
//|      Interval: [Coeff-DeltaCoeff, Coeff+DeltaCoeff]              |
//+------------------------------------------------------------------+

if(idx1 == 0) idx1 = ArraySize(f)-1;
 
 double xavr, yavr, uu, tt, s, sdenom;
   int  n = idx1 - idx0+1;
   
   if( n<=2) return EMPTY_VALUE;
 
   xavr = yavr = 0;   
      for ( int i = idx0; i<=idx1; i++)  { yavr +=  f[i];  xavr +=x[i];  }
   yavr = yavr/n;
   xavr = xavr/n;
 
   s = sdenom = 0;
   for ( int i = 0; i<n; i++)
   {          
      uu = x[i] - xavr;       sdenom = sdenom + uu*uu;
      tt = (f[i] - Coeff*x[i] -B);      s = s + tt*tt;
   }
    

CStudenttDistr StudentDistr; // StudentDistribution
double ConfidenceDeltaCoeff;

   sdenom = sqrt(sdenom);
   uu = sqrt(s/(n-2)) / sdenom ;
   
   tt = StudentDistr.InvStudenttDistribution(n-2,1-alfa*0.5); // critical point for Studentt distribution

       ConfidenceDeltaCoeff = tt*uu;
return ConfidenceDeltaCoeff;   
}